#include <sys/queue.h>
#include <string.h>

#include "spho/spho.h"
#include "spho/scope.h"
#include "spho/bind.h"

struct spho_prebind *
spho_prebind_create(struct spho_scope *sc)
{
	size_t i;
	struct spho_nom *nom;
	struct spho_prebind *bind;

	SPHO_PRECOND(sc != NULL);

	if ((bind = calloc(1, sizeof(*bind))) == NULL) {
		SPHO_ERR(sc->ctx, SPHO_ERR_SYS);
		return (NULL);
	}

	if ((bind->binds = calloc(sc->n_noms, sizeof(*bind->binds))) == NULL) {
		SPHO_ERR(sc->ctx, SPHO_ERR_SYS);
		free(bind);
		return (NULL);
	}

	bind->sc = sc;
	bind->sz = sc->n_noms;


	i = 0;
	SLIST_FOREACH(nom, &sc->noms, next) {
		SPHO_ASSERT(i < bind->sz);
		bind->binds[i++] = (struct spho_prebind_pair) { NULL, 0, NULL };
	}

	return (bind);
}

int
spho_prebind_undef(struct spho_prebind *bind,
    const struct spho_nom *nom)
{
	size_t i;
	SPHO_PRECOND(bind != NULL);
	SPHO_PRECOND(nom != NULL);

	for (i = 0; i < bind->sz; i++) {
		if (nom == bind->binds[i].nom) {
			if (bind->binds[i].kind != SPHO_BIND_UNDECLARED) {
				SPHO_ERR(bind->sc->ctx, SPHO_ERR_BIND_DUPL);
				return (-1);
			}
			bind->binds[i].kind = SPHO_BIND_UNDEFINED;
		}
	}

	return (0);
}
int
spho_prebind_tp(struct spho_prebind *bind, const struct spho_nom *nom,
    const struct spho_tp *tp)
{
	size_t i;

	SPHO_PRECOND(bind != NULL);
	SPHO_PRECOND(nom != NULL);
	SPHO_PRECOND(tp != NULL);

	for (i = 0; i < bind->sz; i++) {
		if (nom == bind->binds[i].nom) {
			if (bind->binds[i].kind != SPHO_BIND_UNDECLARED) {
				SPHO_ERR(bind->sc->ctx, SPHO_ERR_BIND_DUPL);
				return (-1);
			}
			bind->binds[i].kind = SPHO_BIND_TP;
			bind->binds[i].val = (union spho_prebind_val) tp;
			return (0);
		}
	}

	SPHO_ERR(bind->sc->ctx, SPHO_ERR_BIND_NOTFOUND);
	return (-1);
}

int
spho_prebind_tp_op(struct spho_prebind *bind, const struct spho_nom *nom,
    const struct spho_tp_op *op)
{
	size_t i;

	SPHO_PRECOND(bind != NULL);
	SPHO_PRECOND(nom != NULL);
	SPHO_PRECOND(op != NULL);

	for (i = 0; i < bind->sz; i++) {
		if (nom == bind->binds[i].nom) {
			if (bind->binds[i].kind != SPHO_BIND_UNDECLARED) {
				SPHO_ERR(bind->sc->ctx, SPHO_ERR_BIND_DUPL);
				return (-1);
			}
			bind->binds[i].kind = SPHO_BIND_TP_OP;
			bind->binds[i].val = (union spho_prebind_val) op;
			return (0);
		}
	}

	SPHO_ERR(bind->sc->ctx, SPHO_ERR_BIND_NOTFOUND);
	return (-1);
}

int
spho_prebind_member_tp(struct spho_prebind *bind, const struct spho_nom *nom,
    const struct spho_tp *memb_tp)
{
	size_t i;

	SPHO_PRECOND(bind != NULL);
	SPHO_PRECOND(nom != NULL);
	SPHO_PRECOND(memb_tp != NULL);

	for (i = 0; i < bind->sz; i++) {
		if (nom == bind->binds[i].nom) {
			if (bind->binds[i].kind != SPHO_BIND_UNDECLARED) {
				SPHO_ERR(bind->sc->ctx, SPHO_ERR_BIND_DUPL);
				return (-1);
			}
			bind->binds[i].kind = SPHO_BIND_MEMB_TP;
			bind->binds[i].val = (union spho_prebind_val) memb_tp;
			return (0);
		}
	}

	SPHO_ERR(bind->sc->ctx, SPHO_ERR_BIND_NOTFOUND);
	return (-1);
}

struct spho_prebind *
spho_prebind_dupl(const struct spho_prebind *bind)
{
	struct spho_prebind *cp;

	if ((cp = calloc(1, sizeof(*cp))) == NULL) {
		SPHO_ERR(bind->sc->ctx, SPHO_ERR_SYS);
		return (NULL);
	}

	if ((cp->binds = calloc(bind->sz, sizeof(*cp->binds))) == NULL) {
		SPHO_ERR(bind->sc->ctx, SPHO_ERR_SYS);
		free(cp);
		return (NULL);
	}

	memcpy(cp->binds, bind->binds, bind->sz * sizeof(*bind->binds));

	return (cp);
}

void
spho_prebind_free(struct spho_prebind *bind)
{
	free(bind->binds);
	free(bind);
}
